//! `Writeable` is a property to change text of an item

use crate::prelude::*;
use anyhow::Result;

use std::sync::{
    atomic::{AtomicI32, Ordering::SeqCst},
    Arc, RwLock,
};

pub trait ModifyText: Send + Sync {
    fn set_text(&self, text: String) -> Result<()>;
    fn add_letter(&self, letter: char) -> Result<String>;
    fn remove_last(&self) -> Result<Option<String>>;
}

/// `Writeable` gives the ability to modify the text inside an `Textable`
pub struct Writeable {
    gui_handler: Arc<GuiHandler>,

    modifyable: Arc<dyn ModifyText>,
    text_change_executable: Arc<Executable<Option<String>>>,

    activation_changed_executable: RwLock<Option<Box<dyn Fn(bool) -> Result<()> + Send + Sync>>>,

    ui_layer: AtomicI32,
}

impl Writeable {
    /// Factory method for `Writeable`, returns `Arc<Writeable>`.
    pub fn new(
        gui_handler: Arc<GuiHandler>,
        modifyable: Arc<dyn ModifyText>,
        text_change_executable: Arc<Executable<Option<String>>>,
    ) -> Result<Arc<Self>> {
        Ok(Arc::new(Writeable {
            gui_handler,
            modifyable,
            text_change_executable,

            activation_changed_executable: RwLock::new(None),

            ui_layer: AtomicI32::new(0),
        }))
    }

    pub fn set_active(self: &Arc<Self>) -> Result<()> {
        if self.gui_handler.set_active_writeable(self.clone())? {
            if let Some(exec) = &*self.activation_changed_executable.read().unwrap() {
                exec(true)?;
            }
        }

        Ok(())
    }

    pub(crate) fn inactivation_event(&self) -> Result<()> {
        if let Some(exec) = &*self.activation_changed_executable.read().unwrap() {
            exec(false)?;
        }

        Ok(())
    }

    pub fn set_activation_changed<F>(&self, f: F)
    where
        F: Fn(bool) -> Result<()> + Send + Sync + 'static,
    {
        *self.activation_changed_executable.write().unwrap() = Some(Box::new(f));
    }

    /// Add method
    ///
    /// # Arguments
    ///
    /// * `writeable` is a `&Arc<Writeable>` instance that is going to be added
    pub fn add(self: &Arc<Self>) -> Result<()> {
        self.gui_handler
            .add_writeable(self.ui_layer.load(SeqCst), self.clone())
    }

    /// Delete method, has to be explicitly called, otherwise it will remain in memory.
    ///
    /// # Arguments
    ///
    /// * `writeable` is a `&Arc<Writeable>` instance that is going to be deleted
    pub fn delete(self: &Arc<Self>) -> Result<()> {
        self.gui_handler
            .delete_writeable(self.ui_layer.load(SeqCst), self)
    }

    pub fn set_ui_layer(&self, ui_layer: i32) {
        self.ui_layer.store(ui_layer, SeqCst);
    }

    /// Replaces the text of `Textable`
    ///
    /// # Arguments
    ///
    /// * `text` that replaces the current text
    pub fn set_text(&self, text: String) -> Result<()> {
        self.modifyable.set_text(text.clone())?;

        self.text_change_executable
            .execute(if text.is_empty() { None } else { Some(text) })?;

        Ok(())
    }

    /// Adds the letter to Text
    ///
    /// # Arguments
    ///
    /// * `letter` the letter thats going to be added
    pub fn add_letter(&self, letter: char) -> Result<()> {
        self.text_change_executable
            .execute(Some(self.modifyable.add_letter(letter)?))?;

        Ok(())
    }

    /// Removes the last letter
    pub fn remove_last(&self) -> Result<()> {
        self.text_change_executable
            .execute(self.modifyable.remove_last()?)?;

        Ok(())
    }
}
